#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
playbook

Copyright:
Copyright © 2019 Huawei Technologies Co., Ltd. All rights reserved.
"""

import os
import sys
import traceback
import time
import getopt
import codecs
import json
import platform
import logging
from logging.handlers import RotatingFileHandler

is_windows = True if platform.system().lower() == 'windows' else False

if is_windows is False:
    import resource

    # Reset SC_OPEN_MAX, the subprocess.Popen will try to close the maximum
    # number of files after the end of the execution
    # set SC_OPEN_MAX to a reasonable value, to avoid the system's max set
    # to a large value, resulting in high CPU usage.
    # the default value 1024,4096
    try:
        maxfd = os.sysconf("SC_OPEN_MAX")
        if maxfd > 1024:
            resource.setrlimit(resource.RLIMIT_NOFILE, (1024, 4096))
    except:
        pass
        # do nothing

if sys.version_info < (2, 7,):
    PY26 = True
else:
    PY26 = False

if PY26 == False:
    from collections import OrderedDict

from utils import runcontext


class CSCException(Exception):
    pass


class OptionErrorException(CSCException):
    pass


class Playbook:
    def __init__(self):
        # policy root path
        self._root_path = ''
        # log level
        self._log_level = logging.INFO
        self._log_path = None
        self._log_inited = False
        # terminal width
        self._twidth = 0
        # process parameters
        self._current_role = ''
        self._method_name_tags = 'get_tags'
        self._method_name_run = 'run_task'
        self._method_name_title = 'getTitle'

        # count information
        self._count_pass = 0
        self._count_failed = 0
        self._count_error = 0
        self._count_total = 0
        self._count_pass_rule = 0
        self._count_failed_rule = 0

        # load all task python file lists
        self._playbook = os.path.realpath(sys.argv[0])
        paths = os.path.split(self._playbook)
        self._playbookdir = paths[0]
        self._playbookname = paths[1]

        # get command args
        self._action = 'audit'
        self._tags = []
        self._result_lang = 'en_us'
        self._variable_file = ''
        self._extral_value = []
        self._poll_interval = None
        self._simple_result = None
        self._name = ''

        self._result_file = os.path.join(self._playbookdir, 'result.json')

        self._list_tasks = False
        self._result_json = None
        self._tasksdir = ''
        self._varsdir = ''
        self._logger = None
        self._backupdir = ''
        self._tasks = []
        self._templatesdir = ''
        self._runcontext = None

    def _init_log(self):
        self._logger = logging.getLogger('playbook')
        self._logger.setLevel(self._log_level)
        # add file logger
        if self._log_path is None:
            path = os.path.join(self._playbookdir, 'log', 'run.log')
        else:
            path = os.path.join(self._log_path, 'securecat.log')
        f_handler = RotatingFileHandler(path, maxBytes=5 * 1024 * 1024, backupCount=1)
        f_handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter('[%(asctime)s] [%(levelname)s] : %(message)s')

        f_handler.setFormatter(formatter)
        self._logger.addHandler(f_handler)
        # add console logger
        c_handler = logging.StreamHandler(sys.stdout)
        c_handler.setLevel(logging.INFO)
        formatter2 = logging.Formatter('%(message)s')
        c_handler.setFormatter(formatter2)
        self._logger.addHandler(c_handler)
        self._log_inited = True

    def set_role(self, value):
        self._current_role = value

    def parse(self, argv):
        opts, args = getopt.getopt(argv[1:], 'hsa:t:v:i:e:r:l:d:n:',
                                   ['help', 'simple', 'action=', 'tags=',
                                    'variable-file=', 'interval=',
                                    'extral-value=', 'result=', 'result-lang=',
                                    'debug=', 'name=', 'list', 'log-path='])
        for one_option, value in opts:
            if one_option in ('-h', '--help'):
                self.usage()
                sys.exit(0)
            elif one_option in ('-s', '--simple'):
                self._simple_result = True
            elif one_option in ('-a', '--action'):
                if value.lower() in ('audit', 'a'):
                    self._action = 'audit'
                else:
                    raise OptionErrorException('Action type only support audit.')
            elif one_option in ('-t', '--tags'):
                list_value = value.split(',')
                for one_value in list_value:
                    self._tags.append(one_value)
            elif one_option in ('-v', '--variable-file'):
                self._variable_file = value
            elif one_option in ('-e', '--extral-value'):
                self._extral_value.append(value)
            elif one_option in ('-i', '--interval'):
                try:
                    float(value)
                    self._poll_interval = value
                except:
                    raise OptionErrorException('The value of interval should be digit.')
            elif one_option in ('-r', '--result'):
                if os.path.isabs(value) is False:
                    curdir = os.path.realpath(os.curdir)
                    self._result_file = os.path.join(curdir, value)
                else:
                    self._result_file = value
                self._result_file = os.path.realpath(self._result_file)
            elif one_option in ('-l', '--result-lang'):
                if value in ('zh_cn', 'en_us'):
                    self._result_lang = value
                else:
                    raise OptionErrorException('result-lang should be zh_cn or en_us.')
            elif one_option in ('-d', '--debug'):
                value_lower = value.lower()
                if value_lower == 'debug':
                    self._log_level = logging.DEBUG
                elif value_lower == 'info':
                    self._log_level = logging.INFO
                elif value_lower == 'error':
                    self._log_level = logging.ERROR
                else:
                    raise OptionErrorException('result-lang should be debug,info,error.')
            elif one_option in ('--log-path'):
                self._log_path = os.path.realpath(value)
                if os.path.exists(self._log_path) is False:
                    self._log_path = None
            elif one_option in ('-n', '--name'):
                if len(value) > 10:
                    # only keep 10 chars
                    self._name = value[0:10]
                else:
                    self._name = value
            elif one_option in ('--list'):
                self._list_tasks = True
            else:
                raise OptionErrorException('Not support option %s' % one_option)

    def get_tasks(self, task_dir):
        return self.list_tasks_python_files(task_dir, '.py', True)

    def list_tasks_python_files(self, task_dir, wildcard, recursion):
        retfiles = []
        exts = wildcard.split(" ")
        files = os.listdir(task_dir)
        for name in files:
            fullname = os.path.join(task_dir, name)
            if (os.path.isdir(fullname) & recursion):
                retfilestmp = self.list_tasks_python_files(fullname, wildcard, recursion)
                retfiles.extend(retfilestmp)
            else:
                for ext in exts:
                    if (name.endswith(ext)):
                        retfiles.append([task_dir, name])

        return retfiles

    def print_title(self, task_type, msg):

        if msg != '':
            strformat = ('%s [%s] ' % (task_type, msg))
        else:
            strformat = ('%s ' % (task_type))

        leftlen = len(strformat)

        leftlen = 0
        strformat = ('%s{x:*>%s}' % (strformat, leftlen))
        msg = strformat.format(x='')
        self._logger.info(msg)

    def get_module_method_object(self, modulepath, modulename, methodname):

        try:
            sys.path.append(modulepath)
            obj_method = None

            # import the moudle
            obj_module = __import__(modulename)
            if hasattr(obj_module, methodname):
                obj_method = getattr(obj_module, methodname)

            if modulepath in sys.path:
                sys.path.remove(modulepath)

            return obj_method

        except:
            # remove task path from python lib path
            if modulepath in sys.path:
                sys.path.remove(modulepath)
            self._logger.error(traceback.format_exc())
            return None

    def get_ansible_vars(self):
        ansible_vars = {}
        if is_windows:
            ansible_vars['ansible_hostname'] = platform.node()
            ansible_vars['ansible_all_ipv4_addresses'] = ''
            ansible_vars['ansible_distribution'] = platform.system()
            ansible_vars['ansible_distribution_version'] = platform.version()
        else:
            ansible_vars['ansible_hostname'] = platform.system()
            ansible_vars['ansible_all_ipv4_addresses'] = ''
            ansible_vars['ansible_distribution'] = platform.linux_distribution()[0]
            ansible_vars['ansible_distribution_version'] = platform.linux_distribution()[1]

        return ansible_vars

    def usage(self):
        print('')
        print(r'Usage: audit playbook.py')
        print('')
        print('Options:')
        print('  -h                    show this help message and exit')
        print('  --list                show the title and tags of all policies')
        print('  -a ACTION, --action=ACTION')
        print('                        specify action type for task, only support audit now. ')
        print('  -v EXTRA_VARS_FILE, --variable-file=EXTRA_VARS_FILE')
        print('                        specify extra variable file, the file format is json.')
        print('  -i INTERVAL, --interval=INTERVAL')
        print('                        set the interval (in seconds) of internal task polling each other.')
        print('  -e EXTRA_VARS, --extral-value=EXTRA_VARS')
        print('                        set additional variables as key=value.')
        print('  -t TAGS, --tags=TAGS')
        print('                        only run tasks tagged with these values.')
        print('  -r RESULT_FILE, --result=RESULT_FILE')
        print(
            '                        specify result file,default is result.json under the same directory of playbook.')
        print('  -l RESULT_LANG, --result_lang=RESULT_LANG')
        print('                        specify result language(zh_cn or en_us) if benchmark support,default is en_us.')
        print('  -d DEBUG_LEVEL, --debug=DEBUG_LEVEL')
        print('                        specify debug infor level, support debug,info,error')
        print('  -s                    specify whether to generate a condensed result, bypass --result_lang option.')
        print('  -n NAME, --name=NAME')
        print('                        specify extra value append to the name in the result, max length is 10.')

    def list_all_tasks(self):
        # invoke all files
        for t_tasks in self._tasks:
            # add task path to the python lib
            tasklibpath = t_tasks[0]
            (taskname, extension) = os.path.splitext(t_tasks[1])
            try:
                # get tags
                printmsg = ''
                task_method_object = self.get_module_method_object(tasklibpath, taskname, self._method_name_tags)
                if task_method_object is not None:
                    tags = task_method_object()

                task_method_object = self.get_module_method_object(tasklibpath, taskname, self._method_name_title)
                if task_method_object is not None:
                    printmsg = task_method_object(self._result_lang)
                else:
                    printmsg = taskname

                print('%s  %s' % (printmsg, tags))

            except KeyboardInterrupt as e:
                raise e
            except Exception as e:
                # remove task path from python lib path
                if tasklibpath in sys.path:
                    sys.path.remove(tasklibpath)
                self._count_error = self._count_error + 1
                self._count_total = self._count_total + 1
                self._logger.info('  Error')
                self._logger.info('  %s' % e)
                self._logger.error(traceback.format_exc())

    def log_error(self, msg):
        path = os.path.join(self._playbookdir, 'log', 'error.log')
        try:
            logfile = None
            logfile = open(path, 'a')
            if logfile is not None:
                logfile.write(msg)
        except:
            traceback.print_exc()
        finally:
            if logfile:
                logfile.close()

    def set_extra_vars_pre(self):
        pass

    def set_extra_vars_post(self):
        pass

    def run(self):

        try:
            self.parse(sys.argv)
            time.sleep(0.5)

            # delete log/error.log file
            errorlog = os.path.join(self._playbookdir, 'log', 'error.log')
            if os.path.exists(errorlog) is True:
                os.remove(errorlog)

            # Initialize the basic object
            self._runcontext = runcontext.RunContext()
            self._init_log()

            self._runcontext.sleep()

            # if the result file is not in the default path, check if file is exists.
            if self._result_file.startswith(self._playbookdir) == False:
                if os.path.exists(self._result_file):
                    msg = 'The result file <%s> is already exists.' % self._result_file
                    print(msg)
                    exit_code = 1
                    self.log_error(msg)
                    sys.exit(exit_code)

            self._logger.debug('Start run playbook...')
            self.print_title('PLAY', self._playbookname)

            self._tasksdir = os.path.join(self._playbookdir, 'roles', self._current_role, 'tasks')
            self._templatesdir = os.path.join(self._playbookdir, 'roles', self._current_role, 'templates')
            self._varsdir = os.path.join(self._playbookdir, 'roles', self._current_role, 'vars')
            self._backupdir = os.path.join(self._playbookdir, 'roles', self._current_role, 'vars', 'backup')

            if self._result_lang is not None:
                self._runcontext.set_result_lang(self._result_lang)

            if self._simple_result is not None:
                self._runcontext.set_simple_result(self._simple_result)

            if self._poll_interval is not None:
                self._runcontext.set_poll_interval(self._poll_interval)

            # process variables
            self._runcontext.sleep()
            self._logger.debug('Set extra variables pre...')
            self.set_extra_vars_pre()

            self._logger.debug('Process extra variables...')
            if self._variable_file != '':
                if os.path.isabs(self._variable_file) is False:
                    curdir = os.path.realpath(os.curdir)
                    self._variable_file = os.path.join(curdir, self._variable_file)

                self._variable_file = os.path.realpath(self._variable_file)
                if os.path.exists(self._variable_file) is False:
                    raise CSCException('Variable file %s is not exist.' % self._variable_file)
                else:
                    vfp = codecs.open(self._variable_file, 'r', encoding='utf-8')
                    vjson = json.load(vfp)
                    vfp.close()
                    for k in vjson:
                        self._runcontext.set_extra_var(k, vjson[k])

            for ex_value in self._extral_value:
                self._runcontext.set_extra_vars_with_string(ex_value)

            self._logger.debug('Get task list...')
            self._tasks = self.get_tasks(self._tasksdir)

            # process --list option
            if self._list_tasks == True:
                self.list_all_tasks()
                exit_code = 0
                sys.exit(exit_code)

            if self._log_level == logging.DEBUG:
                self._logger.debug('\n'.join([os.path.join(x[0], x[1]) for x in self._tasks]))

            self._runcontext.sleep()
            self._logger.debug('Set extra variables post...')

            self.set_extra_vars_post()

            # print arguments
            tmp_list = []
            tmp_list.append('  Playbook        : %s\n' % self._playbook)
            tmp_list.append('  Name            : %s\n' % self._name)
            tmp_list.append('  Action          : %s\n' % self._action)
            tmp_list.append('  Tags            : %s\n' % self._tags)
            tmp_list.append('  Poll Interval   : %s\n' % self._runcontext.get_poll_interval())
            tmp_list.append('  Result File     : %s\n' % self._result_file)
            tmp_list.append('  Result Lang     : %s\n' % self._runcontext.get_result_lang())
            tmp_list.append('  Variable File   : %s\n' % self._variable_file)
            tmp_list.append('  Extra Variables : %s\n' % self._extral_value)
            tmp_list.append('  All Variables   :\n')
            extra_vars = self._runcontext.get_extra_vars()
            for k in extra_vars:
                tmp_list.append('    %s = %s\n' % (k, extra_vars[k]))

            msg = ''.join(tmp_list)
            self._logger.info(msg)

            self._logger.debug('Init template object...')

            # create Template instance
            self._runcontext.set_template(self._templatesdir)

            # generate total result
            ansible_vars = self.get_ansible_vars()
            ansible_vars['ansible_date_time'] = time.strftime("%b %d %Y %H:%M:%S", time.localtime())

            self._logger.debug('Init result json object...')

            rc_flag, out = self._runcontext.get_template().render(self._runcontext.get_template_file('total.j2'),
                                                                  **ansible_vars)

            self._runcontext.sleep()

            if rc_flag != True:
                exit_code = 2
                self._logger.info('Failed to render total result template.')
                self._logger.debug(out)
                sys.exit(exit_code)
            if PY26:
                self._result_json = json.loads(out)
            else:
                self._result_json = json.loads(out, object_pairs_hook=OrderedDict)

            # reset name
            if self._name != '':
                self._result_json['name'] = '%s-%s' % (self._result_json['type'], self._name)

            self._logger.debug('Set global parameters...')

            # set global parameters
            self._runcontext.set_parameter('playbookdir', self._playbookdir)
            self._runcontext.set_parameter('backup_path', self._backupdir)
            self._runcontext.set_parameter('action', self._action)

            # set audit/harden/rollback module path
            modulepath = os.environ.get('CSC_MODULE_PATH')
            if modulepath is None:
                raise CSCException("Can not find environment variable CSC_MODULE_PATH.")

            self._runcontext.set_module_path(modulepath)

            self._runcontext.sleep()

            self._logger.debug('Run all tasks...')

            # invoke all files
            for t_tasks in self._tasks:
                # add task path to the python lib
                tasklibpath = t_tasks[0]
                (taskname, extension) = os.path.splitext(t_tasks[1])
                try:

                    # get rule tags
                    rule_tags = []
                    task_method_object = self.get_module_method_object(tasklibpath, taskname, self._method_name_tags)
                    if task_method_object is not None:
                        rule_tags = task_method_object()

                    is_in_tag = True
                    if self._tags is not None and len(self._tags) > 0:
                        is_in_tag = False
                        for t_tags in rule_tags:
                            if t_tags in self._tags:
                                is_in_tag = True
                                break

                    if is_in_tag == False:
                        continue

                    rule_tags_len = len(rule_tags)
                    if len(rule_tags) > 0:
                        self.print_title('TASK', rule_tags[rule_tags_len - 1])
                    else:
                        self.print_title('TASK', taskname)

                    task_method_object = self.get_module_method_object(tasklibpath, taskname, self._method_name_run)
                    if task_method_object is not None:
                        rc_flag, out = task_method_object(self._runcontext)

                    r_json = None
                    if rc_flag == 0:
                        if PY26:
                            r_json = json.loads(out)
                        else:
                            r_json = json.loads(out, object_pairs_hook=OrderedDict)
                        if 'title' in r_json:
                            self.print_title(' ', r_json['title'])
                        self._result_json['results'].append(r_json)
                        self._count_pass = self._count_pass + 1
                        if r_json['type'] == 'rule':
                            self._count_pass_rule = self._count_pass_rule + 1
                            self._count_total = self._count_total + 1
                        self._logger.info('  Pass')
                    elif rc_flag == -1:
                        if PY26:
                            r_json = json.loads(out)
                        else:
                            r_json = json.loads(out, object_pairs_hook=OrderedDict)
                        if 'title' in r_json:
                            self.print_title(' ', r_json['title'])
                        self._result_json['results'].append(r_json)
                        self._count_failed = self._count_failed + 1
                        if r_json['type'] == 'rule':
                            self._count_failed_rule = self._count_failed_rule + 1
                            self._count_total = self._count_total + 1
                        self._logger.info('  Failed')
                    else:
                        r_json = {}
                        r_json['msg'] = out
                        self._count_error = self._count_error + 1
                        self._count_total = self._count_total + 1
                        self._logger.info('  Error')
                        self._logger.error(out)

                except KeyboardInterrupt as e:
                    raise e
                except Exception as e:
                    # remove task path from python lib path
                    if tasklibpath in sys.path:
                        sys.path.remove(tasklibpath)
                    self._count_error = self._count_error + 1
                    self._count_total = self._count_total + 1
                    self._logger.info('  Error')
                    self._logger.info('  %s' % e)
                    self._logger.error(traceback.format_exc())

                self._logger.info('\n')
                self._runcontext.sleep()

            resultfp = codecs.open(self._result_file, 'w', encoding='utf-8')
            os.chmod(self._result_file, int('600', 8))
            json.dump(self._result_json, resultfp, ensure_ascii=False, indent=4)
            resultfp.close()

            self.print_title('PLAY RECAP', '')
            self._logger.info('    Total Policies               : %s' % self._count_total)
            self._logger.info('    Total Policies Pass          : %s' % self._count_pass_rule)
            self._logger.info('    Total Policies Failed        : %s' % self._count_failed_rule)
            self._logger.info('    Total Policies Error         : %s' % self._count_error)
            self._logger.info('    Result file: : %s' % self._result_file)
            exit_code = 0
        except OptionErrorException as e:
            print(e)
            self.usage()
            exit_code = 5
            self.log_error('Error Option : ' + str(e))
        except CSCException as e:
            # have been initialize logger sucessful
            print(e)
            if self._log_inited == True:
                self._logger.exception(e)
            exit_code = 1
            self.log_error(traceback.format_exc())
        except KeyboardInterrupt as e:
            print("User interrupted execution.")
            if self._log_inited == True:
                self._logger.exception(e)
            exit_code = 99
            self.log_error("User interrupted execution.")
        except Exception as e:
            print('Unexpected Exception: %s' % e)
            if self._log_inited == True:
                self._logger.exception(e)
            exit_code = 250
            self.log_error(traceback.format_exc())

        sys.exit(exit_code)


if __name__ == '__main__':
    myplaybook = Playbook()
    myplaybook.set_role('hw')
    myplaybook.run()
